@@ -8,6 +8,7 @@ from random import sample,gauss,randrange,uniform
 from scipy.interpolate import LSQUnivariateSpline
 from scipy.signal import savgol_filter
 from scipy.stats import multivariate_normal
+from scipy.special import logsumexp
 
 class Proposal(object):
     """
@@ -259,10 +260,10 @@ class DefaultProposalCycle(ProposalCycle):
                      EnsembleStretch(),
                      DifferentialEvolution(),
                      EnsembleEigenVector()]
-        weights = [2,
-                   2,
-                   5,
-                   1]
+        weights = [3,
+                   3,
+                   1,
+                   10]
         super(DefaultProposalCycle,self).__init__(proposals, weights)
 
 class HamiltonianProposalCycle(ProposalCycle):
@@ -292,9 +293,19 @@ class HamiltonianProposal(EnsembleEigenVector):
         energy and the :obj:`cpnest.Model.potential`.
         """
         super(HamiltonianProposal, self).__init__(**kwargs)
-        self.T  = self.kinetic_energy
-        self.V = model.potential
-        self.normal = None
+        self.T              = self.kinetic_energy
+        self.V              = model.potential
+        self.normal         = None
+        self.dt             = 0.3
+        self.base_dt        = 0.3
+        self.scale          = 1.0
+        self.L              = 20
+        self.base_L         = 20
+        self.TARGET         = 0.500
+        self.ADAPTATIONSIZE = 0.001
+        self._initialised   = False
+        self.c              = self.counter()
+        self.DEBUG          = 0
     
     def set_ensemble(self, ensemble):
         """
@@ -307,7 +318,7 @@ class HamiltonianProposal(EnsembleEigenVector):
         self.update_mass()
         self.update_normal_vector()
         self.update_momenta_distribution()
-
+    
     def update_normal_vector(self):
         """
         update the constraint by approximating the
@@ -325,6 +336,7 @@ class HamiltonianProposal(EnsembleEigenVector):
         self.normal = []
         for i,x in enumerate(tracers_array.T):
             # sort the values
+#            self.normal.append(lambda x: -x)
             idx = x.argsort()
             xs = x[idx]
             Vs = V_vals[idx]
@@ -357,7 +369,7 @@ class HamiltonianProposal(EnsembleEigenVector):
                               )
             # construct a LSQ spline interpolant
             self.normal.append(LSQUnivariateSpline(xs, f, knots, ext = 3, k = 3))
-#            np.savetxt('dlogL_spline_%d.txt'%i,np.column_stack((xs,Vs,self.normal[-1](xs),f)))
+            if self.DEBUG: np.savetxt('dlogL_spline_%d.txt'%i,np.column_stack((xs,Vs,self.normal[-1](xs),f)))
 
     def unit_normal(self, q):
         """
@@ -373,9 +385,9 @@ class HamiltonianProposal(EnsembleEigenVector):
         ----------
         n: :obj:`numpy.ndarray` unit normal to the logLmin contour evaluated at q
         """
-        v = np.array([self.normal[i](q[n]) for i,n in enumerate(q.names)])
-        v[np.isnan(v)] = -1.0
-        n = v/np.linalg.norm(v)
+        v               = np.array([self.normal[i](q[n]) for i,n in enumerate(q.names)])
+        v[np.isnan(v)]  = -1.0
+        n               = v/np.linalg.norm(v)
         return n
 
     def gradient(self, q):
@@ -392,13 +404,13 @@ class HamiltonianProposal(EnsembleEigenVector):
         """
         dV = self.dV(q)
         return dV.view(np.float64)
-    
+        
     def update_momenta_distribution(self):
         """
         update the momenta distribution using the
         mass matrix (precision matrix of the ensemble).
         """
-        self.momenta_distribution = multivariate_normal(cov=self.mass_matrix)
+        self.momenta_distribution = multivariate_normal(cov=self.mass_matrix)#
 
     def update_mass(self):
         """
@@ -406,8 +418,50 @@ class HamiltonianProposal(EnsembleEigenVector):
         inverse mass matrix (precision matrix)
         from the ensemble, allowing for correlated momenta
         """
-        self.mass_matrix = np.linalg.inv(np.atleast_2d(self.covariance))
-        self.inverse_mass_matrix = np.atleast_2d(self.covariance)
+        self.d                      = self.covariance.shape[0]
+        self.inverse_mass_matrix    = np.atleast_2d(self.covariance)
+        self.mass_matrix            = np.linalg.inv(self.inverse_mass_matrix)
+        self.inverse_mass           = np.atleast_1d(np.squeeze(np.diag(self.inverse_mass_matrix)))
+        _, self.logdeterminant      = np.linalg.slogdet(self.mass_matrix)
+        if self._initialised == False:
+            self.set_integration_parameters()
+
+    def set_integration_parameters(self):
+        """
+        Set the integration length according to the N-dimensional ellipsoid
+        shortest and longest principal axes. The former sets to base time step
+        while the latter sets the trajectory length
+        """
+        ranges = [self.prior_bounds[j][1] - self.prior_bounds[j][0] for j in range(self.d)]
+        
+        l, h = np.min(ranges), np.max(ranges)
+        
+        self.base_L         = 10+int((h/l)*self.d**(1./4.))
+        self.base_dt        = (1.0/self.base_L)*l/h
+        self._initialised   = True
+
+
+    def update_time_step(self, acceptance):
+        """
+        Update the time step according to the
+        acceptance rate
+        Parameters
+        ----------
+        acceptance : :obj:'numpy.float'
+        """
+        diff = acceptance - self.TARGET
+        new_log_scale = np.log(self.scale) + self.ADAPTATIONSIZE * diff
+        self.scale = np.exp(new_log_scale)
+        self.dt = self.base_dt * self.scale
+
+    def update_trajectory_length(self,nmcmc):
+        """
+        Update the trajectory length according to the estimated ACL
+        Parameters
+        ----------
+        nmcmc :`obj`:: int
+        """
+        self.L = self.base_L + np.random.randint(nmcmc,5*nmcmc)
 
     def kinetic_energy(self,p):
         """
@@ -421,7 +475,22 @@ class HamiltonianProposal(EnsembleEigenVector):
         ----------
         T: :float: kinetic energy
         """
-        return 0.5 * np.dot(p,np.dot(self.inverse_mass_matrix,p))
+        return 0.5 * np.dot(p,np.dot(self.inverse_mass_matrix,p))-self.logdeterminant-0.5*self.d*np.log(2.0*np.pi)
+
+    def hamiltonian(self, p, q):
+        """
+        Hamiltonian.
+        Parameters
+        ----------
+        p : :obj:`numpy.ndarray`
+            momentum
+        q : :obj:`cpnest.parameter.LivePoint`
+            position
+        Returns
+        ----------
+        H: :float: hamiltonian
+        """
+        return self.T(p) + self.V(q)
 
 class LeapFrog(HamiltonianProposal):
     """
@@ -435,8 +504,8 @@ class LeapFrog(HamiltonianProposal):
         model : :obj:`cpnest.Model`
         """
         super(LeapFrog, self).__init__(model=model, **kwargs)
-        self.dV = model.force
-        self.prior_bounds = model.bounds
+        self.dV             = model.force
+        self.prior_bounds   = model.bounds
 
     def get_sample(self, q0, *args):
         """
@@ -454,17 +523,15 @@ class LeapFrog(HamiltonianProposal):
         """
         # generate a canonical momentum
         p0 = np.atleast_1d(self.momenta_distribution.rvs())
+        initial_energy = self.hamiltonian(p0,q0)
         # evolve along the trajectory
         q, p = self.evolve_trajectory(p0, q0, *args)
-        q0.logP = -self.V(q0)
         # minus sign from the definition of the potential
-        initial_energy = self.T(p0) - q0.logP
-        q.logP = -self.V(q)
-        final_energy   = self.T(p)  - q.logP
-        self.log_J = min(0.0, initial_energy - final_energy)
+        final_energy   = self.hamiltonian(p,q)
+        self.log_J = min(0.0, initial_energy-final_energy)
         return q
     
-    def evolve_trajectory(self, p0, q0):
+    def evolve_trajectory(self, p0, q0, *args):
         """
         Hamiltonian leap frog trajectory subject to the
         hard boundary defined by the parameters prior bounds.
@@ -483,27 +550,24 @@ class LeapFrog(HamiltonianProposal):
         q: :obj:`cpnest.parameter.LivePoint`
             position
         """
-        invM = np.atleast_1d(np.squeeze(np.diag(self.inverse_mass_matrix)))
-        # generate the trajectory lengths from a scale invariant distribution
-        self.L  = int(np.exp(np.random.uniform(np.log(10),np.log(50))))
-        self.dt = 3e-2*float(len(invM))**(-0.25)
-        self.dt = np.abs(gauss(self.dt,self.dt))
         # Updating the momentum a half-step
-        p = p0-0.5 * self.dt * self.gradient(q0)
-        q = q0
+        p = p0 - 0.5 * self.dt * self.gradient(q0)
+        q = q0.copy()
+        
         for i in range(self.L):
             # do a step
             for j,k in enumerate(q.names):
                 u,l = self.prior_bounds[j][1], self.prior_bounds[j][0]
-                q[k] += self.dt * p[j] * invM[j]
+                q[k] += self.dt * p[j] * self.inverse_mass[j]
                 # check and reflect against the bounds
                 # of the allowed parameter range
-                if q[k] > u:
-                    q[k] = u - (q[k] - u)
-                    p[j] *= -1
-                if q[k] < l:
-                    q[k] = l + (l - q[k])
-                    p[j] *= -1
+                while q[k] <= l or q[k] >= u:
+                    if q[k] > u:
+                        q[k] = u - (q[k] - u)
+                        p[j] *= -1
+                    if q[k] < l:
+                        q[k] = l + (l - q[k])
+                        p[j] *= -1
         
             dV = self.gradient(q)
             # take a full momentum step
@@ -544,8 +608,93 @@ class ConstrainedLeapFrog(LeapFrog):
         q: :obj:`cpnest.parameter.LivePoint`
             position
         """
-        return super(ConstrainedLeapFrog,self).get_sample(q0,logLmin)
+        return super(ConstrainedLeapFrog,self).get_sample(q0, logLmin)
     
+    def counter(self):
+        n = 0
+        while True:
+            yield n
+            n += 1
+
+    def evolve_trajectory_one_step_position(self, p, q):
+        """
+        One leap frog step in position
+        
+        Parameters
+        ----------
+        p0 : :obj:`numpy.ndarray`
+            momentum
+        q0 : :obj:`cpnest.parameter.LivePoint`
+            position
+        Returns
+        ----------
+        p: :obj:`numpy.ndarray` updated momentum vector
+        q: :obj:`cpnest.parameter.LivePoint` position
+        """
+        for j,k in enumerate(q.names):
+            u, l  = self.prior_bounds[j][1], self.prior_bounds[j][0]
+            q[k]  += self.dt * p[j] * self.inverse_mass[j]
+            # check and reflect against the bounds
+            # of the allowed parameter range
+            while q[k] < l or q[k] > u:
+                if q[k] > u:
+                    q[k] = u - (q[k] - u)
+                    p[j] *= -1
+                if q[k] < l:
+                    q[k] = l + (l - q[k])
+                    p[j] *= -1
+        return p, q
+
+    def evolve_trajectory_one_step_momentum(self, p, q, logLmin, half = False):
+        """
+        One leap frog step in momentum
+        
+        Parameters
+        ----------
+        p0 : :obj:`numpy.ndarray`
+            momentum
+        q0 : :obj:`cpnest.parameter.LivePoint`
+            position
+        logLmin: :obj:`numpy.float64`
+            loglikelihood constraint
+        Returns
+        ----------
+        p: :obj:`numpy.ndarray` updated momentum vector
+        q: :obj:`cpnest.parameter.LivePoint` position
+        """
+        reflected = 0
+        dV        = self.gradient(q)
+        if half is True:
+            p += - 0.5 * self.dt * dV
+            return p, q, reflected
+        else:
+            c = self.check_constraint(q, logLmin)
+            if c > 0:
+                p += - self.dt * dV
+            else:
+                normal = self.unit_normal(q)
+                p += - 2.0*np.dot(p,normal)*normal
+                reflected = 1
+        return p, q, reflected
+
+    def check_constraint(self, q, logLmin):
+        """
+        Check the likelihood
+        
+        Parameters
+        ----------
+        q0 : :obj:`cpnest.parameter.LivePoint`
+        position
+        logLmin: :obj:`numpy.float64`
+        loglikelihood constraint
+        Returns
+        ----------
+        c: :obj:`numpy.float64` value of the constraint
+        """
+        q.logP  = -self.V(q)
+        q.logL  = self.log_likelihood(q)
+        return q.logL - logLmin
+
     def evolve_trajectory(self, p0, q0, logLmin):
         """
         Evolve point according to Hamiltonian method in
@@ -561,69 +710,62 @@ class ConstrainedLeapFrog(LeapFrog):
         Returns
         ----------
         p: :obj:`numpy.ndarray` updated momentum vector
-        q: :obj:`cpnest.parameter.LivePoint`
-            position
+        q: :obj:`cpnest.parameter.LivePoint` position
         """
-        invM = np.atleast_1d(np.squeeze(np.diag(self.inverse_mass_matrix)))
-        # generate the trajectory lengths from a scale invariant distribution
-        self.L  = int(np.exp(np.random.uniform(np.log(10),np.log(50))))
+
+        trajectory = [(q0,p0)]
+        # evolve forward in time
+        i = 0
+        p, q, reflected = self.evolve_trajectory_one_step_momentum(p0.copy(), q0.copy(), logLmin, half = True)
+        while (i < self.L):
+            p, q            = self.evolve_trajectory_one_step_position(p, q)
+            p, q, reflected = self.evolve_trajectory_one_step_momentum(p, q, logLmin, half = False)
+            trajectory.append((q.copy(),p.copy()))
+            i += 1
+        
+        # evolve backward in time
+        i = 0
+        p, q, reflected = self.evolve_trajectory_one_step_momentum(-p0.copy(), q0.copy(), logLmin, half = True)
+        while (i < self.L):
+            p, q            = self.evolve_trajectory_one_step_position(p, q)
+            p, q, reflected = self.evolve_trajectory_one_step_momentum(p, q, logLmin, half = False)
+            trajectory.append((q.copy(),p.copy()))
+            i += 1
+#            if i == 3*self.L: break
+        p, q, reflected     = self.evolve_trajectory_one_step_momentum(p, q, logLmin, half = True)
+
+        if self.DEBUG: self.save_trajectory(trajectory, logLmin)
+        return self.sample_trajectory(trajectory)
+#        print("dt:",self.dt,"L:",self.L,"actual L:",i,"maxL:",3*self.L)
+#        return trajectory[-1]
+
+    def sample_trajectory(self, trajectory):
         """
-        Step size: 3e-2, manual tuning
-        dimension^-(1/4) based on
-        http://www.homepages.ucl.ac.uk/~ucakabe/papers/Bernoulli_11b.pdf
+        
         """
-        self.dt = 3e-3*float(len(invM))**(-0.25)
-#        f = open("trajectory.txt","w")
-#        for j,k in enumerate(q0.names):
-#            f.write("%s\t"%k)
-#        f.write("barrier\t")
-#        f.write("logL\n")
-        # Updating the momentum a half-step
-#        for j,k in enumerate(q0.names):
-#            f.write("%e\t"%q0[k])
-#        f.write("%e\t"%logLmin)
-#        f.write("%e\n"%q0.logL)
-        # First half-step in momentum
-        p = p0 - 0.5 * self.dt * self.gradient(q0)
-        q = q0
-#        for j,k in enumerate(q.names):
-#            f.write("%e\t"%q[k])
-#        f.write("%e\t"%logLmin)
-#        f.write("%e\n"%q.logL)
-        for i in range(self.L):
-            # do a full step in position
-            for j,k in enumerate(q.names):
-                u,l = self.prior_bounds[j][1], self.prior_bounds[j][0]
-                q[k] += self.dt * p[j] * invM[j]
-                # check and reflect against the bounds
-                # of the allowed parameter range
-                if q[k] > u:
-                    q[k] = u - (q[k] - u)
-                    p[j] *= -1
-                if q[k] < l:
-                    q[k] = l + (l - q[k])
-                    p[j] *= -1
-            dV = self.gradient(q)
-            
-            logL = self.log_likelihood(q)
-            q.logL = logL
-            # if the trajectory led us outside the likelihood bound,
-            # reflect the momentum orthogonally to the surface
-            if np.isfinite(logLmin):
-                
-                if (logL - logLmin) <= 0:
-                    normal = self.unit_normal(q)
-                    p = p - 2.0*np.dot(p,normal)*normal
-                else:
-                    # take a full momentum step
-                    p += - self.dt * dV
-            else:
-                p += - self.dt * dV
-#            for j,k in enumerate(q.names):
-#                f.write("%e\t"%q[k])
-#            f.write("%e\t"%logLmin)
-#            f.write("%e\n"%q.logL)
-        # Do a final update of the momentum for a half step
-        p += - 0.5 * self.dt * dV
-#        f.close()
-        return q, -p
+        logw = np.array([-self.hamiltonian(p,q) for q,p in trajectory[1:-1]])
+        norm = logsumexp(logw)
+        idx  = np.random.choice(range(1,len(trajectory)-1), p = np.exp(logw  - norm))
+        return trajectory[idx]
+                    
+    def save_trajectory(self, trajectory, logLmin, filename = None):
+        """
+        save trajectory for diagnostic purposes
+        """
+        if filename is None:
+            filename = 'trajectory_'+str(next(self.c))+'.txt'
+        f = open(filename,'w')
+        names = trajectory[0][0].names
+
+        for n in names:
+            f.write(n+'\t'+'p_'+n+'\t')
+        f.write('logPrior\tlogL\tlogLmin\n')
+
+        for j,step in enumerate(trajectory):
+            q = step[0]
+            p = step[1]
+            for j,n in enumerate(names):
+                f.write(repr(q[n])+'\t'+repr(p[j])+'\t')
+            f.write(repr(q.logP)+'\t'+repr(q.logL)+'\t'+repr(logLmin)+'\n')
+        f.close()
+        if self.c == 3: exit()
